"""Test the metrics linked to the power profile."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from os.path import dirname, join

import pandas as pd
import pytest

from skcycling import Rider
from skcycling.metrics import aerobic_meta_model

module_path = dirname(__file__)
filename_csv = join(module_path, 'data', 'rider_power_profile.csv')
rider = Rider.from_csv(filename_csv)


@pytest.mark.parametrize(
    "rpp, ts, expected_mpa, expected_time_mpa, expected_aei",
    [(rider.record_power_profile()['power'], None,
      453.372, pd.Timedelta('00:03:00'), -11.48848),
     (rider.record_power_profile()['power'],
      pd.timedelta_range('00:00:01', '04:00:00', freq='5S'),
      444.0816, pd.Timedelta('00:03:16'), -10.99544),
     (rider.record_power_profile()['power'],
      pd.timedelta_range('00:00:01', '05:00:00', freq='5S'),
      444.0816, pd.Timedelta('00:03:16'), -10.99544)])
def test_aerobic_meta_model(rpp, ts,
                            expected_mpa,
                            expected_time_mpa,
                            expected_aei):
    mpa, time_mpa, aei, _, _ = aerobic_meta_model(rpp, ts)
    print(time_mpa)
    assert mpa == pytest.approx(expected_mpa)
    assert time_mpa == expected_time_mpa
    assert aei == pytest.approx(expected_aei)



"""Test the power package which model power using aside data."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import pytest

import numpy as np
from numpy.testing import assert_array_less
from numpy.testing import assert_allclose

import pandas as pd
from pandas.testing import assert_series_equal

from skcycling.model import strava_power_model
from skcycling.extraction import gradient_elevation
from skcycling.extraction import acceleration
from skcycling.exceptions import MissingDataError


speed = np.ones(100) * 5
distance = np.linspace(0, 99, num=100)
elevation = np.linspace(0, 9, num=100)
activity = pd.DataFrame({'speed': speed,
                         'distance': distance,
                         'elevation': elevation})


@pytest.mark.parametrize(
    "activity_corrupted, use_acceleration",
    [(activity.copy().drop(columns='distance'), False),
     (activity.copy().drop(columns='speed'), True)]
)
def test_strava_power_model_error(activity_corrupted, use_acceleration):
    with pytest.raises(MissingDataError):
        strava_power_model(activity_corrupted, cyclist_weight=70,
                           use_acceleration=use_acceleration)


def test_strava_power_model_auto_compute():
    # check that the acceleration and the elevation will be auto-computed
    power_auto = strava_power_model(activity, cyclist_weight=70)

    activity_ele_acc = activity.copy()
    activity_ele_acc = gradient_elevation(activity)
    activity_ele_acc = acceleration(activity_ele_acc)
    power_ele_acc = strava_power_model(activity_ele_acc, cyclist_weight=70)

    assert_series_equal(power_auto, power_ele_acc)


def test_strava_power_model():
    # at constant speed the acceleration should not have any influence
    power_without_acc = strava_power_model(activity, cyclist_weight=78,
                                           use_acceleration=False)
    power_with_acc = strava_power_model(activity, cyclist_weight=78,
                                        use_acceleration=True)
    assert_allclose(power_without_acc, power_with_acc)

    # increase cyclist weight should increase power
    power_initial = strava_power_model(activity, cyclist_weight=70)
    power_increase_weight = strava_power_model(activity, cyclist_weight=78)
    assert_array_less(power_initial, power_increase_weight)

    # increase bike weight should increase power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       bike_weight=7)
    power_increase_weight = strava_power_model(activity, cyclist_weight=70,
                                               bike_weight=8)
    assert_array_less(power_initial, power_increase_weight)

    # increase the rolling coefficient should increase power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       coef_roll_res=0.0045)
    power_increase_cr = strava_power_model(activity, cyclist_weight=70,
                                           coef_roll_res=0.006)
    assert_array_less(power_initial, power_increase_cr)

    # increase of the pressure should increase power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       pressure=101325)
    power_increase_pressure = strava_power_model(activity, cyclist_weight=70,
                                                 pressure=110000)
    assert_array_less(power_initial, power_increase_pressure)

    # decrease the temperature should increase the power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       temperature=15.0)
    power_decrease_temperature = strava_power_model(activity,
                                                    cyclist_weight=70,
                                                    temperature=10.0)
    assert_array_less(power_initial, power_decrease_temperature)

    # increase the drag coefficient should increase the power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       coef_drag=0.5)
    power_increase_coef_drag = strava_power_model(activity,
                                                  cyclist_weight=70,
                                                  coef_drag=1.0)
    assert_array_less(power_initial, power_increase_coef_drag)

    # increase the rider frontal surface should increase the power
    power_initial = strava_power_model(activity, cyclist_weight=70,
                                       surface_rider=0.32)
    power_increase_surface_rider = strava_power_model(activity,
                                                      cyclist_weight=70,
                                                      surface_rider=0.5)
    assert_array_less(power_initial, power_increase_surface_rider)



# coding: utf-8

""" Metrics to asses the performance of a cycling ride.

Functions named as ``*_score`` return a scalar value to maximize: the higher
the better.

Function named as ``*_error`` or ``*_loss`` return a scalar value to minimize:
the lower the better.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from __future__ import division

import numpy as np

TS_SCALE_GRAPPE = dict([('I1', 2.), ('I2', 2.5), ('I3', 3.),
                        ('I4', 3.5), ('I5', 4.5), ('I6', 7.),
                        ('I7', 11.)])

ESIE_SCALE_GRAPPE = dict([('I1', (.3, .5)), ('I2', (.5, .6)),
                          ('I3', (.6, .75)), ('I4', (.75, .85)),
                          ('I5', (.85, 1.)), ('I6', (1., 1.80)),
                          ('I7', (1.8, 3.))])


def mpa2ftp(mpa):
    """Convert the maximum power aerobic into the functional threshold power.

    Parameters
    ----------
    mpa : float
        Maximum power aerobic.

    Return:
    -------
    ftp : float
        Functional threshold power.

    Examples
    --------
    >>> from skcycling.metrics import mpa2ftp
    >>> print(mpa2ftp(400)) # doctest: +ELLIPSIS
    304...

    """
    return 0.76 * mpa


def ftp2mpa(ftp):
    """Convert the functional threshold power into the maximum threshold power.

    Parameters
    ----------
    ftp : float
        Functional threshold power.

    Return:
    -------
    mpa : float
        Maximum power aerobic.

    Examples
    --------
    >>> from skcycling.metrics import ftp2mpa
    >>> print(ftp2mpa(304)) # doctest: +ELLIPSIS
    400...

    """
    return ftp / 0.76


def normalized_power_score(activity_power, mpa, window_width=30):
    """Normalized power®.

    The normalized power is an average power computing a smoothed power input
    and rejecting the low power intensities.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    activity_power : Series
        A Series containing the power data from an activity.

    mpa : float
        Maximum power aerobic. Use :func:`metrics.ftp2mpa` if you use the
        functional threshold power metric.


    window_width : int, optional
        The width of the window used to smooth the power data before to compute
        the normalized power. The default width is 30 samples.

    Returns
    -------
    score : float
        Normalized power score.

    References
    ----------
    .. [1] Allen, H., and A. Coggan. "Training and racing with a power
       meter." VeloPress, 2012.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.metrics import normalized_power_score
    >>> ride = bikeread(load_fit()[0])
    >>> mpa = 400
    >>> np = normalized_power_score(ride['power'], mpa)
    >>> print('Normalized power {:.2f} W'.format(np))
    Normalized power 218.49 W

    """

    smooth_activity = (activity_power.rolling(window_width, center=True)
                                     .mean().dropna())
    # removing value < I1-ESIE, i.e. 30 % MPA
    smooth_activity = smooth_activity[
        smooth_activity > ESIE_SCALE_GRAPPE['I1'][0] * mpa]

    return (smooth_activity ** 4).mean() ** (1 / 4)


def intensity_factor_score(activity_power, mpa):
    """Intensity factor®.

    The intensity factor® is the ratio of the normalized power® over the
    functional threshold power. Note that all our computation consider the
    maximum power aerobic for consistency. If you only have the functional
    threshold power, use :func:`metrics.ftp2mpa`.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    activity_power : Series
        A Series containing the power data from an activity.

    mpa : float
        Maximum power aerobic. Use :func:`metrics.ftp2mpa` if you use the
        functional threshold power metric.

    Returns
    -------
    score: float
        Intensity factor.

    References
    ----------
    .. [1] Allen, H., and A. Coggan. "Training and racing with a power
       meter." VeloPress, 2012.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.metrics import intensity_factor_score
    >>> ride = bikeread(load_fit()[0])
    >>> mpa = 400
    >>> if_score = intensity_factor_score(ride['power'], mpa)
    >>> print('Intensity factor {:.2f} W'.format(if_score))
    Intensity factor 0.72 W

    """
    ftp = mpa2ftp(mpa)
    return normalized_power_score(activity_power, mpa) / ftp


def training_stress_score(activity_power, mpa):
    """Training stress score®.

    The training stress score® corresponds to the intensity factor® normalized
    by the time of the activity. You can use the function
    :func:`metrics.ftp2mpa` if you are using the functional threshold metric.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    activity_power : Series
        A Series containing the power data from an activity.

    mpa : float
        Maximum power aerobic. Use :func:`metrics.ftp2mpa` if you use the
        functional threshold power metric.

    Returns
    -------
    score: float
        Training stress score.

    References
    ----------
    .. [1] Allen, H., and A. Coggan. "Training and racing with a power
       meter." VeloPress, 2012.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.metrics import training_stress_score
    >>> ride = bikeread(load_fit()[0])
    >>> mpa = 400
    >>> ts_score = training_stress_score(ride['power'], mpa)
    >>> print('Training stress score {:.2f}'.format(ts_score))
    Training stress score 32.38

    """
    activity_power = activity_power.resample('1S').mean()
    if_score = intensity_factor_score(activity_power, mpa)
    return (activity_power.size * if_score ** 2) / 3600 * 100


def training_load_score(activity_power, mpa):
    """Training load score.

    Grappe et al. proposes to compute the load of an activity by a weighted sum
    of the time spend in the different ESIE zones.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    activity_power : Series
        A Series containing the power data from an activity.

    mpa : float
        Maximum power aerobic. Use :func:`metrics.ftp2mpa` if you use the
        functional threshold power metric.

    Returns
    -------
    tls_score: float
        Training load score.

    References
    ----------
    .. [1] Grappe, F. "Cyclisme et optimisation de la performance: science
       et méthodologie de l'entraînement." De Boeck Supérieur, 2009.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.metrics import training_load_score
    >>> ride = bikeread(load_fit()[0])
    >>> mpa = 400
    >>> tl_score = training_load_score(ride['power'], mpa)
    >>> print('Training load score {:.2f}'.format(tl_score))
    Training load score 74.90

    """
    tls_score = 0.
    activity_power = activity_power.resample('1S').mean()
    for key in TS_SCALE_GRAPPE.keys():
        power_samples = activity_power[
            np.bitwise_and(activity_power >= ESIE_SCALE_GRAPPE[key][0] * mpa,
                           activity_power < ESIE_SCALE_GRAPPE[key][1] * mpa)]
        tls_score += power_samples.size / 60 * TS_SCALE_GRAPPE[key]
    return tls_score



"""
The :mod:`skcycling.datasets` module includes utilities to load datasets.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from os import listdir
from os.path import dirname, join

__all__ = ['load_fit',
           'load_rider']


def load_fit(returned_type='list_file', set_data='normal'):
    """Return path to some FIT toy data.

    Read more in the :ref:`User Guide <datasets>`.

    Parameters
    ----------
    returned_type : str, optional (default='list_file')
        If 'list_file', return a list containing the fit files;
        If 'path', return a string where the data are localized.

    set_data : str, optional (default='normal')
        If 'normal', return 3 files.
        If 'corrupted, return corrupted files for testing.

    Returns
    -------
    filenames : str or list of str,
        List of string or string depending of input parameters.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> load_fit() # doctest : +ELLIPSIS
    [...]

    """
    module_path = dirname(__file__)

    if set_data == 'normal':
        if returned_type == 'list_file':
            return sorted([
                join(module_path, 'data', name)
                for name in listdir(join(module_path, 'data'))
                if name.endswith('.fit')
            ])
        elif returned_type == 'path':
            return join(module_path, 'data')
    elif set_data == 'corrupted':
        if returned_type == 'list_file':
            return sorted([
                join(module_path, 'corrupted_data', name)
                for name in listdir(
                    join(module_path, 'corrupted_data'))
                if name.endswith('.fit')
            ])
        elif returned_type == 'path':
            return join(module_path, 'corrupted_data')


def load_rider():
    """Return the path to a CSV file containing rider information.

    Read more in the :ref:`User Guide <datasets>`.

    Parameters
    ----------
    None

    Returns
    -------
    filename : str
        The path to the CSV file.

    Examples
    --------
    >>> from skcycling.datasets import load_rider
    >>> load_rider() # doctest : +ELLIPSIS
    '...rider.csv'

    """
    module_path = dirname(__file__)

    return join(module_path, 'data', 'rider.csv')



"""
The :mod:`skcycling.model` module includes algorithms to model cycling data.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from .power import strava_power_model

__all__ = ['strava_power_model']



PACKAGE_NAME = 'skcycling'


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(PACKAGE_NAME, parent_package, top_path)

    config.add_subpackage('__check_build')

    # pure python packages
    config.add_subpackage('datasets')
    config.add_subpackage('datasets/tests')
    config.add_subpackage('io')
    config.add_subpackage('io/tests')
    config.add_subpackage('metrics')
    config.add_subpackage('metrics/tests')
    config.add_subpackage('utils')
    config.add_subpackage('utils/tests')

    # packages that have their own setup.py -> cython files
    config.add_subpackage('extraction')

    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



"""Utilities for input validation."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clauses

import glob
import os
from itertools import chain


def validate_filenames(filenames):
    """Check the filenames and expand in the case of wildcard.

    Parameters
    ----------
    filenames : str or list of str

        The path to the file to read. The input accepted are:

        * a filename or a list of filename to the file to read;
        * a filename or a list of filename containing a wildcard
          (e.g. ``'./data/*.fit'``).

    Returns
    -------
    filenames : list of str
        Returns a list of all file names.

    Examples
    --------
    >>> from skcycling.utils import validate_filenames
    >>> from skcycling.datasets import load_fit
    >>> filenames = validate_filenames(load_fit())
    >>> list(filenames) # doctest : +ELLIPSIS
    [...]

    """
    if isinstance(filenames, list):
        return chain.from_iterable([sorted(glob.glob(os.path.expanduser(f)))
                                    for f in filenames])
    else:
        return sorted(glob.glob(os.path.expanduser(filenames)))



"""
The :mod:`skcycling.utils` module include utility functions.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from .validation import validate_filenames


__all__ = ['validate_filenames']



"""
The :mod:`skcycling.io` module includes utility to load data.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from .base import bikeread

__all__ = ['bikeread']



"""Base classes for data management."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import numpy as np
import pandas as pd

from .extraction import activity_power_profile
from .io import bikeread
from .utils import validate_filenames


class Rider(object):
    """User interface for a rider.

    User interface to easily add, remove, compute information related to power.

    Read more in the :ref:`User Guide <record_power_profile>`.

    Parameters
    ----------
    n_jobs : int, (default=1)
        The number of workers to use for the different processing.

    Attributes
    ----------
    power_profile_ : DataFrame
        DataFrame containing all information regarding the power-profile of a
        rider for each ride.

    """

    def __init__(self, n_jobs=1):
        self.n_jobs = n_jobs
        self.power_profile_ = None

    def add_activities(self, filenames):
        """Compute the power-profile for each activity and add it to the
        current power-profile.

        Parameters
        ----------
        filenames : str or list of str
            A string a list of string to the file to read. You can use
            wildcards to automatically check several files.

        Returns
        -------
        None

        Examples
        --------
        >>> from skcycling.datasets import load_fit
        >>> from skcycling.base import Rider
        >>> rider = Rider()
        >>> rider.add_activities(load_fit()[0])
        >>> rider.power_profile_.head()
                          2014-05-07 12:26:22
        cadence 00:00:01            78.000000
                00:00:02            64.000000
                00:00:03            62.666667
                00:00:04            62.500000
                00:00:05            64.400000

        """
        filenames = validate_filenames(filenames)
        activities_pp = [activity_power_profile(bikeread(f))
                         for f in filenames]
        activities_pp = pd.concat(activities_pp, axis=1)

        if self.power_profile_ is not None:
            try:
                self.power_profile_ = self.power_profile_.join(activities_pp,
                                                               how='outer')
            except ValueError as e:
                if 'columns overlap but no suffix specified' in e.args[0]:
                    raise ValueError('One of the activity was already added'
                                     ' to the rider power-profile. Remove this'
                                     ' activity before to try to add it.')
                else:
                    raise
        else:
            self.power_profile_ = activities_pp

    def delete_activities(self, dates, time_comparison=False):
        """Delete the activities power-profile from some specific dates.

        Parameters
        ----------
        dates : list/tuple of datetime-like or str
            The dates of the activities to be removed. The format expected is:

            * datetime-like or str: a single activity will be deleted.
            * a list of datetime-like or str: each activity for which the date
              is contained in the list will be deleted.
            * a tuple of datetime-like or str ``(start_date, end_date)``: the
              activities for which the dates are included in the range will be
              deleted.

        time_comparison : bool, optional
            Whether to make a strict comparison using time or to relax to
            constraints with only the date.

        Returns
        -------
        None

        Examples
        --------
        >>> from skcycling.datasets import load_rider
        >>> from skcycling import Rider
        >>> rider = Rider.from_csv(load_rider())
        >>> rider.delete_activities('07 May 2014')
        >>> print(rider)
        RIDER INFORMATION:
         power-profile:
                           2014-05-11 09:39:38  2014-07-26 16:50:56
        cadence 00:00:01           100.000000            60.000000
                00:00:02            89.000000            58.000000
                00:00:03            68.333333            56.333333
                00:00:04            59.500000            59.250000
                00:00:05            63.200000            61.000000

        """
        def _strict_comparison(dates_pp, date, strict_equal):
            if strict_equal:
                return dates_pp == date
            else:
                return np.bitwise_and(
                    dates_pp >= date,
                    dates_pp <= pd.Timestamp(date) + pd.DateOffset(1))

        if isinstance(dates, tuple):
            if len(dates) != 2:
                raise ValueError("Wrong tuple format. Expecting a tuple of"
                                 " format (start_date, end_date). Got {!r}"
                                 " instead.".format(dates))
            mask_date = np.bitwise_and(
                self.power_profile_.columns >= dates[0],
                self.power_profile_.columns <= pd.Timestamp(dates[1]) +
                pd.DateOffset(1))
        elif isinstance(dates, list):
            mask_date = np.any(
                [_strict_comparison(self.power_profile_.columns, d,
                                    time_comparison)
                 for d in dates], axis=0)
        else:
            mask_date = _strict_comparison(self.power_profile_.columns, dates,
                                           time_comparison)

        mask_date = np.bitwise_not(mask_date)
        self.power_profile_ = self.power_profile_.loc[:, mask_date]

    def record_power_profile(self, range_dates=None, columns=None):
        """Compute the record power-profile.

        Parameters
        ----------
        range_dates : tuple of datetime-like or str, optional
            The start and end date to consider when computing the record
            power-profile. By default, all data will be used.

        columns : array-like or None, optional
            Name of data field to return. By default, all available data will
            be returned.

        Returns
        -------
        record_power_profile : DataFrame
            Record power-profile taken between the range of dates.

        Examples
        --------
        >>> from skcycling import Rider
        >>> from skcycling.datasets import load_rider
        >>> rider = Rider.from_csv(load_rider())
        >>> record_power_profile = rider.record_power_profile()
        >>> record_power_profile.head() # doctest: +NORMALIZE_WHITESPACE
                    cadence      distance  elevation  heart-rate       power
        00:00:01  60.000000  27162.600000        NaN         NaN  750.000000
        00:00:02  58.000000  27163.750000        NaN         NaN  741.000000
        00:00:03  56.333333  27164.586667        NaN         NaN  731.666667
        00:00:04  59.250000  27163.402500        NaN         NaN  719.500000
        00:00:05  61.000000  27162.142000        NaN         NaN  712.200000

        This is also possible to give a range of dates to compute the record
        power-profile. We can also select some specific information.

        >>> record_power_profile = rider.record_power_profile(
        ...     range_dates=('07 May 2014', '11 May 2014'),
        ...     columns=['power', 'cadence'])
        >>> record_power_profile.head()
                     cadence   power
        00:00:01  100.000000  717.00
        00:00:02   89.000000  717.00
        00:00:03   68.333333  590.00
        00:00:04   59.500000  552.25
        00:00:05   63.200000  552.60

        """
        if range_dates is None:
            mask_date = np.ones_like(self.power_profile_.columns,
                                     dtype=bool)
        else:
            mask_date = np.bitwise_and(
                self.power_profile_.columns >= range_dates[0],
                self.power_profile_.columns <= pd.Timestamp(range_dates[1]) +
                pd.DateOffset(1))

        if columns is None:
            columns = self.power_profile_.index.levels[0]

        pp_idxmax = (self.power_profile_.loc['power']
                                        .loc[:, mask_date]
                                        .idxmax(axis=1)
                                        .dropna())
        rpp = {}
        for dt in columns:
            data = self.power_profile_.loc[dt].loc[:, mask_date]
            rpp[dt] = pd.Series(
                [data.loc[date_idx]
                 for date_idx in pp_idxmax.iteritems()],
                index=data.index[:pp_idxmax.size])

        return pd.DataFrame(rpp)

    @classmethod
    def from_csv(cls, filename, n_jobs=1):
        """Load rider information from a CSV file.

        Parameters
        ----------
        filename : str
            The path to the CSV file.

        n_jobs : int, (default=1)
            The number of workers to use for the different processing.

        Returns
        -------
        rider : skcycling.Rider
            The :class:`skcycling.Rider` instance.

        Examples
        --------
        >>> from skcycling.datasets import load_rider
        >>> from skcycling import Rider
        >>> rider = Rider.from_csv(load_rider())
        >>> print(rider) # doctest: +NORMALIZE_WHITESPACE
        RIDER INFORMATION:
         power-profile:
                           2014-05-07 12:26:22  2014-05-11 09:39:38  \\
        cadence 00:00:01            78.000000           100.000000
                00:00:02            64.000000            89.000000
                00:00:03            62.666667            68.333333
                00:00:04            62.500000            59.500000
                00:00:05            64.400000            63.200000
        <BLANKLINE>
                          2014-07-26 16:50:56
        cadence 00:00:01            60.000000
                00:00:02            58.000000
                00:00:03            56.333333
                00:00:04            59.250000
                00:00:05            61.000000

        """
        df = pd.read_csv(filename, index_col=[0, 1])
        df.columns = pd.to_datetime(df.columns)
        df.index = pd.MultiIndex(levels=[df.index.levels[0],
                                         pd.to_timedelta(df.index.levels[1])],
                                 labels=df.index.labels,
                                 name=[None, None])
        rider = cls(n_jobs=n_jobs)
        rider.power_profile_ = df
        return rider

    def to_csv(self, filename):
        """Drop the rider information into a CSV file.

        Parameters
        ----------
        filename : str
            The path to the CSV file.

        Returns
        -------
        None

        Examples
        --------
        >>> from skcycling.datasets import load_fit
        >>> from skcycling import Rider
        >>> rider = Rider(n_jobs=-1)
        >>> rider.add_activities(load_fit()[:1])
        >>> print(rider)
        RIDER INFORMATION:
         power-profile:
                           2014-05-07 12:26:22
        cadence 00:00:01            78.000000
                00:00:02            64.000000
                00:00:03            62.666667
                00:00:04            62.500000
                00:00:05            64.400000

        """
        self.power_profile_.to_csv(filename, date_format='%Y-%m-%d %H:%M:%S')

    def __repr__(self):
        return 'RIDER INFORMATION:\n power-profile:\n {}'.format(
            self.power_profile_.head())



"""Methods to handle input/output files."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import os
from collections import defaultdict

import pandas as pd
import numpy as np
import six

from fitparse import FitFile

# 'timestamp' will be consider as the index of the DataFrame later on
FIELDS_DATA = ('timestamp', 'power', 'heart_rate', 'cadence', 'distance',
               'altitude', 'speed')


def check_filename_fit(filename):
    """Method to check if the filename corresponds to a fit file.

    Parameters
    ----------
    filename : str
        The fit file to check.

    Returns
    -------
    filename : str
        The checked filename.

    """

    # Check that filename is of string type
    if isinstance(filename, six.string_types):
        # Check that this is a fit file
        if filename.endswith('.fit'):
            # Check that the file is existing
            if os.path.isfile(filename):
                return filename
            else:
                raise ValueError('The file does not exist.')
        else:
            raise ValueError('The file is not a fit file.')
    else:
        raise ValueError('filename needs to be a string. Got {}'.format(
            type(filename)))


def load_power_from_fit(filename):
    """Method to open the power data from FIT file into a pandas dataframe.

    Parameters
    ----------
    filename : str,
        Path to the FIT file.

    Returns
    -------
    data : DataFrame
        Power records of the ride.

    """
    filename = check_filename_fit(filename)
    activity = FitFile(filename)
    activity.parse()
    records = activity.get_messages(name='record')

    data = defaultdict(list)
    for rec in records:
        values = rec.get_values()
        for key in FIELDS_DATA:
            data[key].append(values.get(key, np.NaN))

    data = pd.DataFrame(data)
    if data.empty:
        raise IOError('The file {} does not contain any data.'.format(
            filename))

    # rename the columns for consistency
    data.rename(columns={'heart_rate': 'heart-rate', 'altitude': 'elevation'},
                inplace=True)

    data.set_index(FIELDS_DATA[0], inplace=True)
    del data.index.name

    return data



"""
The :mod:`skcycling.metrics` module include score functions.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from .activity import normalized_power_score
from .activity import intensity_factor_score
from .activity import training_stress_score
from .activity import training_load_score
from .activity import mpa2ftp
from .activity import ftp2mpa

from .power_profile import aerobic_meta_model

__all__ = ['normalized_power_score',
           'intensity_factor_score',
           'training_stress_score',
           'training_load_score',
           'mpa2ftp',
           'ftp2mpa',
           'aerobic_meta_model']



"""
The :mod:`skcycling.extraction` module includes algorithms to extract
information from cycling data.
"""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from .gradient import acceleration
from .gradient import gradient_activity
from .gradient import gradient_elevation
from .gradient import gradient_heart_rate

from .power_profile import activity_power_profile


__all__ = ['acceleration',
           'gradient_activity',
           'gradient_elevation',
           'gradient_heart_rate',
           'activity_power_profile']



"""Methods to load power data file."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import numpy as np

from .fit import load_power_from_fit

DROP_OPTIONS = ('columns', 'rows', 'both')


def bikeread(filename, drop_nan=None):
    """Read power data file.

    Read more in the :ref:`User Guide <reader>`.

    Parameters
    ----------
    filename : str
        Path to the file to read.

    drop_nan : str {'columns', 'rows', 'both'} or None
        Either to remove the columns/rows containing NaN values. By default,
        all data will be kept.

    Returns
    -------
    data : DataFrame
        Power data and time data.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> activity = bikeread(load_fit()[0], drop_nan='columns')
    >>> activity.head() # doctest : +NORMALIZE_WHITESPACE
                         elevation  cadence  distance  power  speed
    2014-05-07 12:26:22       64.8     45.0      3.05  256.0  3.036
    2014-05-07 12:26:23       64.8     42.0      6.09  185.0  3.053
    2014-05-07 12:26:24       64.8     44.0      9.09  343.0  3.004
    2014-05-07 12:26:25       64.8     45.0     11.94  344.0  2.846
    2014-05-07 12:26:26       65.8     48.0     15.03  389.0  3.088

    """
    if drop_nan is not None and drop_nan not in DROP_OPTIONS:
        raise ValueError('"drop_nan" should be one of {}.'
                         ' Got {} instead.'.format(DROP_OPTIONS, drop_nan))

    df = load_power_from_fit(filename)

    if drop_nan is not None:
        if drop_nan == 'columns':
            df.dropna(axis=1, inplace=True)
        elif drop_nan == 'rows':
            df.dropna(axis=0, inplace=True)
        else:
            df.dropna(axis=1, inplace=True).dropna(axis=0, inplace=True)

    # remove possible outliers by clipping the value
    df[df['power'] > 2500.] = np.nan

    # resample to have a precision of a second with additional linear
    # interpolation for missing value
    return df.resample('s').interpolate('linear')



# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import os
import shutil
from tempfile import mkdtemp

import pytest
from pandas.testing import assert_frame_equal

from skcycling.base import Rider
from skcycling.datasets import load_fit
from skcycling.datasets import load_rider


def test_rider_add_activities_update():
    rider = Rider.from_csv(load_rider())
    rider.delete_activities('07 May 2014')
    rider.add_activities(load_fit()[0])
    assert rider.power_profile_.shape == (35771, 3)

    with pytest.raises(ValueError, message='activity was already added'):
        rider.add_activities(load_fit()[0])


@pytest.mark.parametrize(
    "rider, filename, expected_shape",
    [(Rider(), load_fit(), (40218, 3)),
     (Rider(), load_fit()[0], (13536, 1))])
def test_rider_add_activities(rider, filename, expected_shape):
    rider.add_activities(filename)
    assert rider.power_profile_.shape == expected_shape


@pytest.mark.parametrize(
    "dates, time_comparison, expected_shape",
    [('07 May 2014', False, (33515, 2)),
     ('07 May 2014', True, (33515, 3)),
     ('07 May 2014 12:26:22', True, (33515, 2)),
     (['07 May 2014'], False, (33515, 2)),
     (tuple(['07 May 2014', '11 May 2014']), False, (33515, 1))])
def test_rider_delete_activities(dates, time_comparison, expected_shape):
    rider = Rider.from_csv(load_rider())
    rider.delete_activities(dates, time_comparison=time_comparison)
    assert rider.power_profile_.shape == expected_shape


@pytest.mark.parametrize(
    "dates",
    [(tuple(['07 May 2014'])),
     (tuple(['07 May 2014', '10 May 2014', '11 May 2014']))])
def test_rider_delete_activities_error(dates):
    rider = Rider.from_csv(load_rider())

    msg = "Wrong tuple format"
    with pytest.raises(ValueError, message=msg):
        rider.delete_activities(dates)


@pytest.mark.parametrize(
    "range_dates, expected_shape",
    [(None, (6703, 5)),
     (('07 May 2014', '11 May 2014'), (3812, 5))])
def test_rider_record_power_profile(range_dates, expected_shape):
    rider = Rider.from_csv(load_rider())
    rpp = rider.record_power_profile(range_dates=range_dates)
    assert rpp.shape == expected_shape


def test_dump_load_rider():
    filenames = load_fit()[:1]
    rider = Rider()
    rider.add_activities(filenames)

    tmpdir = mkdtemp()
    csv_filename = os.path.join(tmpdir, 'rider.csv')
    try:
        rider.to_csv(csv_filename)
        rider2 = Rider.from_csv(csv_filename)
        assert_frame_equal(rider.power_profile_, rider2.power_profile_)
    finally:
        shutil.rmtree(tmpdir)



# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from os.path import dirname, join

import pytest

from skcycling.datasets import load_fit
from skcycling.utils import validate_filenames

filenames = load_fit()


@pytest.mark.parametrize(
    "filenames, expected_filenames",
    [(filenames, filenames),
     (join(dirname(filenames[0]), '*.fit'), filenames)])
def test_validate_filenames(filenames, expected_filenames):
    assert list(validate_filenames(filenames)) == expected_filenames



# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from os.path import join

import pytest

from skcycling.datasets import load_fit
from skcycling.datasets import load_rider


@pytest.mark.parametrize(
    "returned_type, set_data, expected_filenames",
    [('list_file', 'normal', sorted(['2014-05-11-11-39-38.fit',
                                     '2014-05-07-14-26-22.fit',
                                     '2014-07-26-18-50-56.fit'])),
     ('path', 'normal', ['data']),
     ('list_file', 'corrupted', sorted(['2013-04-24-22-22-25.fit',
                                        '2014-05-17-10-44-53.fit',
                                        '2015-11-27-18-54-57.fit'])),
     ('path', 'corrupted', ['corrupted_data'])])
def test_load_fit(returned_type, set_data, expected_filenames):
    filenames = load_fit(returned_type=returned_type,
                         set_data=set_data)
    if not isinstance(filenames, list):
        filenames = [filenames]
    for f, e in zip(filenames, expected_filenames):
        assert e in f


def test_load_rider():
    filename = load_rider()
    assert join('data', 'rider.csv') in filename



"""Function to extract gradient information about different features."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from __future__ import division

from collections import Iterable

import pandas as pd

from ..exceptions import MissingDataError


def acceleration(activity, periods=5, append=True):
    """Compute the acceleration (i.e. speed gradient).

    Read more in the :ref:`User Guide <gradient>`.

    Parameters
    ----------
    activity : DataFrame
        The activity containing speed information.

    periods : int, default=5
        Periods to shift to compute the acceleration.

    append : bool, optional
        Whether to append the acceleration to the original activity (default)
        or to only return the acceleration as a Series.

    Returns
    -------
    data : DataFrame or Series
        The original activity with an additional column containing the
        acceleration or a single Series containing the acceleration.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.extraction import acceleration
    >>> ride = bikeread(load_fit()[0])
    >>> new_ride = acceleration(ride)

    """
    if 'speed' not in activity.columns:
        raise MissingDataError('To compute the acceleration, speed data are '
                               'required. Got {} fields.'
                               .format(activity.columns))

    acceleration = activity['speed'].diff(periods=periods) / periods

    if append:
        activity['acceleration'] = acceleration
        return activity
    else:
        return acceleration


def gradient_elevation(activity, periods=5, append=True):
    """Compute the elevation gradient.

    Read more in the :ref:`User Guide <gradient>`.

    Parameters
    ----------
    activity : DataFrame
        The activity containing elevation and distance information.

    periods : int, default=5
        Periods to shift to compute the elevation gradient.

    append : bool, optional
        Whether to append the elevation gradient to the original activity
        (default) or to only return the elevation gradient as a Series.

    Returns
    -------
    data : DataFrame or Series
        The original activity with an additional column containing the
        elevation gradient or a single Series containing the elevation
        gradient.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.extraction import gradient_elevation
    >>> ride = bikeread(load_fit()[0])
    >>> new_ride = gradient_elevation(ride)

    """
    if not {'elevation', 'distance'}.issubset(activity.columns):
        raise MissingDataError('To compute the elevation gradient, elevation '
                               'and distance data are required. Got {} fields.'
                               .format(activity.columns))

    diff_elevation = activity['elevation'].diff(periods=periods)
    diff_distance = activity['distance'].diff(periods=periods)
    gradient_elevation = diff_elevation / diff_distance

    if append:
        activity['gradient-elevation'] = gradient_elevation
        return activity
    else:
        return gradient_elevation


def gradient_heart_rate(activity, periods=5, append=True):
    """Compute the heart-rate gradient.

    Read more in the :ref:`User Guide <gradient>`.

    Parameters
    ----------
    activity : DataFrame
        The activity containing heart-rate information.

    periods : int, default=5
        Periods to shift to compute the heart-rate gradient.

    append : bool, optional
        Whether to append the heart-rate gradient to the original activity
        (default) or to only return the heart-rate gradient as a Series.

    Returns
    -------
    data : DataFrame or Series
        The original activity with an additional column containing the
        heart-rate gradient or a single Series containing the heart-rate
        gradient.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.extraction import gradient_heart_rate
    >>> ride = bikeread(load_fit()[0])
    >>> ride['heart-rate'] = pd.Series(
    ...     np.random.randint(60, 200, size=ride.shape[0]),
    ...     index=ride.index)  # Add fake heart-rate data for the example
    >>> new_ride = gradient_heart_rate(ride)

    """
    if 'heart-rate' not in activity.columns:
        raise MissingDataError('To compute the heart-rate gradient, heart-rate'
                               ' data are required. Got {} fields.'
                               .format(activity.columns))

    gradient_heart_rate = activity['heart-rate'].diff(periods=periods)

    if append:
        activity['gradient-heart-rate'] = gradient_heart_rate
        return activity
    else:
        return gradient_heart_rate


def gradient_activity(activity, periods=1, append=True, columns=None):
    """Compute the gradient for all given columns.

    Read more in the :ref:`User Guide <gradient>`.

    Parameters
    ----------
    activity : DataFrame
        The activity to use to compute the gradient.

    periods : int or array-like, default=1
        Periods to shift to compute the gradient. If an array-like is given,
        several gradient will be computed.

    append : bool, optional
        Whether to append the gradients to the original activity.

    columns : list, optional
        The name of the columns to use to compute the gradient. By default, all
        the columns are used.

    Returns
    -------
    gradient : DataFrame
        The computed gradient from the activity.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.extraction import gradient_activity
    >>> ride = bikeread(load_fit()[0], drop_nan='columns')
    >>> new_ride = acceleration(ride)

    """
    if columns is not None:
        data = activity[columns]
    else:
        data = activity

    if isinstance(periods, Iterable):
        gradient = [data.diff(periods=p) for p in periods]
        gradient_name = ['gradient_{}'.format(p) for p in periods]
    else:
        gradient = [data.diff(periods=periods)]
        gradient_name = ['gradient_{}'.format(periods)]

    if append:
        # prepend the original information
        gradient = [activity] + gradient
        gradient_name = ['original'] + gradient_name

    return pd.concat(gradient, axis=1, keys=gradient_name)



"""Module to estimate power from data."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from __future__ import division

import numpy as np
from scipy import constants

from ..extraction import gradient_elevation
from ..extraction import acceleration


def strava_power_model(activity, cyclist_weight, bike_weight=6.8,
                       coef_roll_res=0.0045, pressure=101325.0,
                       temperature=15.0, coef_drag=1, surface_rider=0.32,
                       use_acceleration=False):
    """Strava model used to estimate power.

    It corresponds the mathematical formulation which add all forces applied to
    a cyclist in movement.

    Read more in the :ref:`User Guide <strava>`.

    Parameters
    ----------
    activity : DataFrame
        The activity containing the ride information.

    cyclist_weight : float
        The cyclist weight in kg.

    bike_weight : float, default=6.8
        The bike weight in kg.

    coef_roll_res : float, default=0.0045
        Rolling resistance coefficient.

    pressure : float, default=101325.0
        Pressure in Pascal.

    temperature : float, default=15.0
        Temperature in Celsius.

    coef_drag : float, default=1
        The drag coefficient also known as Cx.

    surface_rider : float, default=0.32
        Surface area of the rider facing wind also known as S. The unit is m^2.

    use_acceleration : bool, default=False
        Either to add the power required to accelerate. This estimation can
        become unstable if the acceleration varies for reason which are not
        linked to power changes (i.e., braking, bends, etc.)

    Returns
    -------
    power : Series
        The power estimated.

    References
    ----------
    .. [1] How Strava Calculates Power
       https://support.strava.com/hc/en-us/articles/216917107-How-Strava-Calculates-Power

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.model import strava_power_model
    >>> ride = bikeread(load_fit()[0])
    >>> power = strava_power_model(ride, cyclist_weight=72)
    >>> print(power['2014-05-07 12:26:28':
    ...             '2014-05-07 12:26:38'])  # Show 10 sec of estimated power
    2014-05-07 12:26:28    196.567898
    2014-05-07 12:26:29    198.638094
    2014-05-07 12:26:30    191.444894
    2014-05-07 12:26:31     26.365864
    2014-05-07 12:26:32     89.826104
    2014-05-07 12:26:33    150.842325
    2014-05-07 12:26:34    210.083958
    2014-05-07 12:26:35    331.573965
    2014-05-07 12:26:36    425.013711
    2014-05-07 12:26:37    428.806914
    2014-05-07 12:26:38    425.410451
    Freq: S, dtype: float64

    """
    if 'gradient-elevation' not in activity.columns:
        activity = gradient_elevation(activity)
    if use_acceleration and 'acceleration' not in activity.columns:
        activity = acceleration(activity)

    temperature_kelvin = constants.convert_temperature(
        temperature, 'Celsius', 'Kelvin')
    total_weight = cyclist_weight + bike_weight  # kg

    speed = activity['speed']  # m.s^-1
    power_roll_res = coef_roll_res * constants.g * total_weight * speed

    # air density at 0 degree Celsius and a standard atmosphere
    molar_mass_dry_air = 28.97 / 1000  # kg.mol^-1
    standard_atmosphere = constants.physical_constants[
        'standard atmosphere'][0]  # Pa
    zero_celsius_kelvin = constants.convert_temperature(
        0, 'Celsius', 'Kelvin')  # 273.15 K
    air_density_ref = (
        (standard_atmosphere * molar_mass_dry_air) /
        (constants.gas_constant * zero_celsius_kelvin))  # kg.m^-3
    air_density = air_density_ref * (
        (pressure * zero_celsius_kelvin) /
        (standard_atmosphere * temperature_kelvin))  # kg.m^-3
    power_wind = 0.5 * air_density * surface_rider * coef_drag * speed**3

    slope = activity['gradient-elevation']  # grade
    power_gravity = (total_weight * constants.g *
                     np.sin(np.arctan(slope)) * speed)

    power_total = power_roll_res + power_wind + power_gravity

    if use_acceleration:
        acc = activity['acceleration']  # m.s^-1
        power_acceleration = total_weight * acc * speed
        power_total = power_total + power_acceleration

    return power_total.clip(0)



""" Metrics to asses the power profile. """

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from __future__ import division

import pandas as pd
import numpy as np

from sklearn.linear_model import LinearRegression


SAMPLING_WKO = pd.TimedeltaIndex(
    ['00:00:01', '00:00:05', '00:00:30', '00:01:00', '00:03:00',
     '00:03:30', '00:04:00', '00:04:30', '00:05:00', '00:05:30',
     '00:06:00', '00:06:30', '00:07:00', '00:10:00', '00:20:00',
     '00:30:00', '00:45:00', '01:00:00', '02:00:00', '03:00:00',
     '04:00:00'])


def std_dev_squared_error(y_true, y_pred):
    """Compute the standard deviation of the squared error.

    Parameters
    ----------
    y_true : ndarray, shape (n_samples,)
        Ground truth (correct) target values.

    y_pred : ndarray, shape (n_samples,)
        Estimated target values.

    Returns
    -------
    std_dev : float
        Standard deviation of the squared error.

    """

    return np.sqrt(np.sum((y_true - y_pred) ** 2 / (y_true.size - 2)))


def aerobic_meta_model(record_power_profile, time_samples=None):
    """Compute the aerobic metabolism model from the record power-profile.

    Read more in the :ref:`User Guide <mpa_estimate>`.

    Parameters
    ----------
    record_power_profile : Series
        The record power profile from which to extract the aerobic model.

    time_samples : TimedeltaIndex or None, optional
        The time samples of the record power-profile to take into account. If
        None, the sampling of the method of Pinot et al. is applied, which is
        equivalent to the sampling from WKO+.

    Returns
    -------
    mpa : float
        Maximum Aerobic Power.

    t_mpa : Timedelta
        Time of the Maximum Aerobic Power.

    aei : float
        Aerobic Endurance Index.

    fit_info_mpa_fitting : dict
        This is a dictionary with the information collected about the fitting
        related to the MAP. The attributes will be the following:

        - `slope`: slope of the linear fitting,
        - `intercept`: intercept of the linear fitting,
        - `std_err`: standard error of the fitting,
        - `coeff_det`: coefficient of determination.

    fit_info_aei_fitting : dict
        This is a dictionary with the information collected about the fitting
        related to the AEI. The attributes will be the following:

        - `slope`: slope of the linear fitting,
        - `intercept`: intercept of the linear fitting,
        - `std_err`: standard error of the fitting,
        - `coeff_det`: coefficient of determination.

    Notes
    -----
    The method implemented here follow the work presented in [1]_.

    References
    ----------
    .. [1] Pinot et al., "Determination of Maximal Aerobic Power
       on the Field in Cycling", Jounal of Science and Cycling, vol. 3(1),
       pp. 26-31, 2014.

    """
    if time_samples is None:
        time_samples = SAMPLING_WKO.copy()

    # keep only the time samples available in the record power-profile
    mask_time_samples = time_samples < record_power_profile.index.max()
    time_samples = time_samples[mask_time_samples]

    # to avoid losing data, we will first interpolate the time samples
    # using all the data available in the record power-profile before
    # to select only the samples required.
    ts_union = record_power_profile.index.union(time_samples)
    record_power_profile = (record_power_profile.reindex(ts_union)
                                                .interpolate('linear')
                                                .reindex(time_samples))

    # only samples between 10 minutes and 4 hours are considered for the
    # regression
    mask_samples_map = np.bitwise_and(time_samples >= '00:10:00',
                                      time_samples <= '04:00:00')
    extracted_profile = record_power_profile.loc[mask_samples_map].values
    extracted_time = record_power_profile.loc[mask_samples_map].index.values
    extracted_time = np.log(extracted_time /
                            np.timedelta64(1, 's')).reshape(-1, 1)

    ols = LinearRegression()
    ols.fit(extracted_time, extracted_profile)
    std_fit = std_dev_squared_error(extracted_profile,
                                    ols.predict(extracted_time))

    fit_info_mpa_fitting = {
        'slope': ols.coef_[0],
        'intercept': ols.intercept_,
        'std_err': std_fit,
        'coeff_det': ols.score(extracted_time, extracted_profile)}

    # mpa will be find between 3 minutes and 7 minutes
    mask_samples_map = np.bitwise_and(time_samples >= '00:03:00',
                                      time_samples <= '00:10:00')
    extracted_profile = record_power_profile.loc[mask_samples_map].values
    extracted_time = record_power_profile.loc[mask_samples_map].index.values
    extracted_time = np.log(extracted_time /
                            np.timedelta64(1, 's')).reshape(-1, 1)
    aerobic_model = ols.predict(extracted_time)

    # find the first value in the 2 * std confidence interval
    samples_within = np.abs(extracted_profile - aerobic_model) < 2 * std_fit

    if np.count_nonzero(samples_within):
        index_mpa = np.flatnonzero(samples_within)[0]
        time_mpa = record_power_profile.loc[mask_samples_map].index[index_mpa]
        mpa = record_power_profile.loc[mask_samples_map].iloc[index_mpa]
    else:
        raise ValueError('There is no value entering in the confidence'
                         ' level between 3 and 7 minutes.')

    # find aerobic endurance index
    mask_samples_aei = np.bitwise_and(time_samples >= time_mpa,
                                      time_samples <= '04:00:00')
    extracted_profile = record_power_profile.loc[mask_samples_aei].values
    extracted_profile = extracted_profile / mpa * 100
    extracted_time = record_power_profile.loc[mask_samples_aei].index.values
    extracted_time = np.log(extracted_time /
                            np.timedelta64(1, 's')).reshape(-1, 1)

    ols.fit(extracted_time, extracted_profile)
    fit_info_aei_fitting = {
        'slope': ols.coef_[0],
        'intercept': ols.intercept_,
        'std_err': std_fit,
        'coeff_det': ols.score(extracted_time, extracted_profile)}

    return (mpa, time_mpa, ols.coef_[0],
            fit_info_mpa_fitting, fit_info_aei_fitting)



"""Cycling Processing Toolbox (Toolbox for SciPy)

``scikit-cycling`` (a.k.a ``skcycling``) is a set of python methods to
analyse file extracted from powermeters.

Subpackages
-----------
datasets
    Modules with helper for datasets.
metrics
    Metrics to quantify cyclist ride.
power_profile
    Record power-profile of cyclist.
utils
    Utility to read and save cycling ride.
"""

from ._version import __version__

from . import __check_build
from .base import Rider



#! /usr/bin/env python
"""A set of python modules for cyclist using powermeters."""

import io
import re
import os
import sys
import subprocess
import codecs

from setuptools import find_packages

PACKAGE_NAME = 'skcycling'
DISTNAME = 'scikit-cycling'
DESCRIPTION = 'A set of python modules for cyclist using powermeters.'
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = 'G. Lemaitre'
MAINTAINER_EMAIL = 'g.lemaitre58@gmail.com'
URL = 'https://github.com/scikit-cycling/scikit-cycling'
LICENSE = 'BSD3'
DOWNLOAD_URL = 'https://github.com/scikit-cycling/scikit-cycling'
INSTALL_REQUIRES = ['numpy', 'scipy', 'pandas', 'scikit-learn', 'six',
                    'joblib', 'fitparse', 'cython']
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'License :: OSI Approved',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6']
PACKAGE_DATA = {'skcycling': ['datasets/data/*.fit',
                              'datasets/data/*.csv',
                              'datasets/corrupted_data/*.fit',
                              'metrics/tests/data/rider_power_profile.csv']}
EXTRAS_REQUIRE = {
    'tests': [
        'pytest',
        'pytest-cov'],
    'docs': [
        'sphinx',
        'sphinx-gallery',
        'sphinx_rtd_theme',
        'numpydoc',
        'matplotlib',
    ]
}


def version(package, encoding='utf-8'):
    """Obtain the packge version from a python file e.g. pkg/_version.py
    See <https://packaging.python.org/en/latest/single_source_version.html>.
    """
    path = os.path.join(os.path.dirname(__file__), package, '_version.py')
    with io.open(path, encoding=encoding) as fp:
        version_info = fp.read()
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_info, re.M)
    if not version_match:
        raise RuntimeError("Unable to find version string.")
    return version_match.group(1)


def generate_cython(package):
    """Cythonize all sources in the package"""
    cwd = os.path.abspath(os.path.dirname(__file__))
    print("Cythonizing sources")
    p = subprocess.call([sys.executable,
                         os.path.join(cwd, 'tools', 'cythonize.py'),
                         package],
                        cwd=cwd)
    if p != 0:
        raise RuntimeError("Running cythonize failed!")


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage(PACKAGE_NAME)

    return config


def setup_package():
    from numpy.distutils.core import setup

    old_path = os.getcwd()
    local_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    src_path = local_path

    os.chdir(local_path)
    sys.path.insert(0, local_path)

    # Run build
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    cwd = os.path.abspath(os.path.dirname(__file__))
    if not os.path.exists(os.path.join(cwd, 'PKG-INFO')):
        # Generate Cython sources, unless building from source release
        generate_cython(PACKAGE_NAME)

    try:
        setup(name=DISTNAME,
              author=MAINTAINER,
              author_email=MAINTAINER_EMAIL,
              maintainer=MAINTAINER,
              maintainer_email=MAINTAINER_EMAIL,
              description=DESCRIPTION,
              license=LICENSE,
              url=URL,
              version=version(PACKAGE_NAME),
              download_url=DOWNLOAD_URL,
              long_description=LONG_DESCRIPTION,
              zip_safe=False,  # the package can run out of an .egg file
              classifiers=CLASSIFIERS,
              packages=find_packages(),
              package_data=PACKAGE_DATA,
              install_requires=INSTALL_REQUIRES,
              extras_require=EXTRAS_REQUIRE,
              configuration=configuration)
    finally:
        del sys.path[0]
        os.chdir(old_path)

    return


if __name__ == '__main__':
    setup_package()



import numpy


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('extraction', parent_package, top_path)
    libraries = []
    config.add_extension('_power_profile',
                         sources=['_power_profile.c'],
                         include_dirs=[numpy.get_include()],
                         libraries=libraries,
                         extra_compile_args=["-O3", "-fopenmp"],
                         extra_link_args=["-fopenmp"])
    config.add_subpackage("tests")

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup
    setup(**configuration().todict())



"""Module containing the exceptions used in scikit-cycling."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

__all__ = ['MissingDataError']


class MissingDataError(ValueError):
    """Error raised when there is not the required data to make some
    computation.

    For instance, :func:`skcycling.extraction.gradient_elevation` required
    elevation and distance data which might not be provided. In this case, this
    type of error is raised.

    """



"""Extraction of information based on the power-profile."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

from datetime import time, timedelta
from numbers import Integral

import numpy as np
import pandas as pd

from ._power_profile import max_mean_power_interval
from ._power_profile import _associated_data_power_profile


def activity_power_profile(activity, max_duration=None):
    """Compute the power profile for an activity.

    Read more in the :ref:`User Guide <activity_power_profile>`.

    Parameters
    ----------
    activity : DataFrame
        A pandas DataFrame with at least a ``'power'`` column and the indices
        are the information about time. The activity can be read with
        :func:`skcycling.io.bikeread`.

    max_duration : Timedelta, timedelta, np.timedelta64, int, or str, optional
        The maximum duration for which the power-profile should be computed. By
        default, it will be computed for the duration of the activity. An
        integer represents seconds.

    Returns
    -------
    power_profile : Series
        A pandas Series containing the power-profile.

    References
    ----------
    .. [1] Pinot, J., and F. Grappe. "The record power profile to assess
       performance in elite cyclists." International journal of sports medicine
       32.11 (2011): 839-844.

    Examples
    --------
    >>> from skcycling.datasets import load_fit
    >>> from skcycling.io import bikeread
    >>> from skcycling.extraction import activity_power_profile
    >>> power_profile = activity_power_profile(bikeread(load_fit()[0]))
    >>> power_profile.head() # doctest : +NORMALIZE_WHITESPACE
    cadence  00:00:01    78.000000
             00:00:02    64.000000
             00:00:03    62.666667
             00:00:04    62.500000
             00:00:05    64.400000
    Name: 2014-05-07 12:26:22, dtype: float64

    """
    if max_duration is None:
        max_duration = pd.Timedelta(seconds=activity.shape[0])
    elif isinstance(max_duration, Integral):
        max_duration = pd.Timedelta(seconds=max_duration)
    else:
        max_duration = pd.Timedelta(max_duration)

    max_duration = min(
        max_duration,
        activity.index[-1] - activity.index[0] + pd.Timedelta(seconds=1))

    activity_power = activity['power']
    activity_complement = activity.drop(['power'], axis=1)

    # use the threading backend since we release the GIL.
    power_profile, power_profile_idx = zip(
        *[max_mean_power_interval(activity_power.values, duration)
          for duration in range(1, max_duration.seconds)])
    power_profile = np.array(power_profile)
    power_profile_idx = np.array(power_profile_idx)

    series_index = pd.timedelta_range(
        "00:00:01", timedelta(seconds=max_duration.seconds - 1), freq='s')
    series_name = pd.Timestamp(activity.index[0])

    # if some additional data are available, we will add them as them on the
    # side of the power-profile.
    if not activity_complement.empty:
        complement_data = {col: pd.Series(
            _associated_data_power_profile(activity_complement[col].values,
                                           power_profile_idx,
                                           np.arange(1, max_duration.seconds,
                                                     dtype=int)),
            index=series_index, name=series_name)
                           for col in activity_complement.columns}
        complement_data['power'] = pd.Series(power_profile, index=series_index,
                                             name=series_name)
        return pd.concat(complement_data)

    else:
        return pd.Series(power_profile, index=series_index, name=series_name)



"""Testing the metrics developed to asses performance of a ride."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Cedric Lemaitre
# License: BSD 3 clause

import pytest

import pandas as pd
import numpy as np

from skcycling.metrics import normalized_power_score
from skcycling.metrics import intensity_factor_score
from skcycling.metrics import training_stress_score
from skcycling.metrics import training_load_score
from skcycling.metrics import mpa2ftp
from skcycling.metrics import ftp2mpa


mpa = 400.
ftp = 304.

ride = np.array([300.] * 200 + [0.] * 200 + [200.] * 200)
ride = pd.Series(ride,
                 index=pd.date_range('1/1/2011',
                                     periods=ride.size,
                                     freq='1S'),
                 name='power')

ride_2 = np.array([140.] * 20 + [220.] * 20 + [250.] * 20 + [310.] * 20 +
                  [350.] * 20 + [410.] * 20 + [800.] * 20)
ride_2 = pd.Series(ride_2,
                   index=pd.date_range('1/1/2011',
                                       periods=ride_2.size,
                                       freq='1S'),
                   name='power')


@pytest.mark.parametrize(
    "score_func, params, expected_score",
    [(normalized_power_score, (ride, mpa), 260.7611),
     (intensity_factor_score, (ride, mpa), 0.857766),
     (training_stress_score, (ride, mpa), 12.26273),
     (training_load_score, (ride_2, mpa), 11.16666)]
)
def test_scores(score_func, params, expected_score):
    assert score_func(*params) == pytest.approx(expected_score)


def test_convert_mpa_ftp():
    assert mpa2ftp(ftp2mpa(ftp)) == pytest.approx(ftp)
